import requests
from bs4 import BeautifulSoup
import pandas as pd
from datetime import datetime
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D


# Function to fetch the webpage's HTML content
def fetch_page_data(url):
    response = requests.get(url)  # Send an HTTP GET request to the provided URL
    soup = BeautifulSoup(response.text, 'html.parser')  # Parse the page content using BeautifulSoup
    return soup


# Function to extract data from the table based on FOSS filters and date range
def extract_table_data(soup, foss_filter, start_date, end_date):
    table = soup.find('table', {'class': 'paleblue table table-striped table-hover table-bordered'})  # Find the desired table by class name
    if not table:
        return []  # If no table is found, return an empty list

    rows = table.find_all('tr')[1:]  # Skip the header row and get all the other rows
    data = []
    for row in rows:
        cols = row.find_all('td')[:-1]  # Extract columns (skip last column)
        row_data = [col.text.strip() for col in cols]  # Extract text from each column and strip whitespace
        if row_data[4] in foss_filter:  # Apply filter based on FOSS type
            date = datetime.strptime(row_data[8], '%d %b %Y')  # Convert date string to datetime object
            if start_date <= date <= end_date:  # Check if the date is within the given range
                data.append(row_data)  # Add row data to the list
    return data


# Function to find the total number of pages in the pagination
def find_total_pages(soup):
    pagination = soup.find('ul', {'class': 'pagination'})  # Locate the pagination element
    if pagination:
        pages = [link.text.strip() for link in pagination.find_all('a') if link.text.strip().isdigit()]  # Extract page numbers
        if pages:
            return int(pages[-1])  # Return the last page number as the total number of pages
    return 1  # If no pagination is found, assume there's only one page


# Function to scrape data across all pages
def scrape_all_pages(base_url, foss_filter, start_date, end_date):
    initial_soup = fetch_page_data(base_url)  # Fetch the first page
    total_pages = find_total_pages(initial_soup)  # Get the total number of pages
    all_data = extract_table_data(initial_soup, foss_filter, start_date, end_date)  # Extract data from the first page

    # Loop through each subsequent page and scrape data
    for page in range(2, total_pages + 1):
        url = f"{base_url}&page={page}"  # Construct the URL for the next page
        soup = fetch_page_data(url)  # Fetch the page content
        page_data = extract_table_data(soup, foss_filter, start_date, end_date)  # Extract data from the page
        all_data.extend(page_data)  # Add the data to the overall list

    return all_data  # Return all the scraped data


# Function to generate a pie chart visualization of FOSS categories
def piechart_visualization(df, writer):
    foss_counts = df['FOSS'].value_counts()  # Count occurrences of each FOSS category

    # Create the pie chart
    fig, ax = plt.subplots(figsize=(8, 6))
    foss_counts.plot(kind='pie', autopct='%1.1f%%', startangle=90, colors=['lightcoral', 'lightblue', 'lightgreen', 'gold'], ax=ax)
    plt.title('FOSS Category Distribution (Pie Chart)')
    plt.ylabel('')  # Hide the y-axis label
    plt.tight_layout()  # Adjust layout
    plt.show()

    # Write the FOSS counts to an Excel sheet
    df_sheet_name = 'FOSS Visualization'
    df_foss_counts = pd.DataFrame(foss_counts).reset_index()
    df_foss_counts.columns = ['FOSS Category', 'Number of Workshops']
    df_foss_counts.to_excel(writer, sheet_name=df_sheet_name, index=False)

    return foss_counts  # Return the FOSS counts


# Function to generate a bar chart visualization of city-wise workshops
def barchart_visualization(df, writer):
    city_counts = df['City'].value_counts()  # Count workshops per city

    # Create the bar chart
    fig, ax = plt.subplots(figsize=(8, 6))
    city_counts.plot(kind='bar', color='lightgreen', ax=ax)
    plt.title('City-wise Workshop Distribution (Bar Chart)')
    plt.xlabel('City')
    plt.ylabel('Number of Workshops')
    plt.tight_layout()
    plt.show()

    # Write the city counts to an Excel sheet
    df_sheet_name = 'City Visualization'
    df_city_counts = pd.DataFrame(city_counts).reset_index()
    df_city_counts.columns = ['City', 'Number of Workshops']
    df_city_counts.to_excel(writer, sheet_name=df_sheet_name, index=False)

    return city_counts  # Return the city counts


# Function to analyze specific data for a college
def analyze_college_data(df, college_name):
    college_data = df[df['Institution'].str.contains(college_name, case=False, na=False)]  # Filter data for the specific college

    if college_data.empty:
        return None, None, None  # Return None if no data is found for the college

    unique_foss = college_data['FOSS'].nunique()  # Count unique FOSS categories
    foss_types = college_data['FOSS'].unique()  # Get the unique FOSS types

    unique_departments = college_data['Department'].nunique()  # Count unique departments
    departments = college_data['Department'].unique()  # Get the unique departments

    unique_organisers = college_data['Organiser'].nunique()  # Count unique organizers
    organisers = college_data['Organiser'].unique()  # Get the unique organizers

    return foss_types, departments, organisers  # Return the relevant data


# Function to generate a 3D visualization of workshops by city, FOSS type, and participants
def three_d_visualization(df):
    df['Participants'] = pd.to_numeric(df['Participants'], errors='coerce').fillna(0)  # Convert participant count to numeric and handle errors
    city_foss_participants = df.groupby(['City', 'FOSS'])['Participants'].sum().reset_index()  # Group by city and FOSS type

    cities = city_foss_participants['City'].unique()  # Get unique cities
    foss_types = city_foss_participants['FOSS'].unique()  # Get unique FOSS types

    city_map = {city: i for i, city in enumerate(cities)}  # Map city names to numeric values for the chart
    foss_map = {foss: i for i, foss in enumerate(foss_types)}  # Map FOSS types to numeric values

    city_foss_participants['City_num'] = city_foss_participants['City'].map(city_map)  # Apply city mapping
    city_foss_participants['FOSS_num'] = city_foss_participants['FOSS'].map(foss_map)  # Apply FOSS mapping

    # Create the 3D bar chart
    fig = plt.figure(figsize=(10, 8))
    ax = fig.add_subplot(111, projection='3d')

    ax.bar3d(
        city_foss_participants['City_num'],
        city_foss_participants['FOSS_num'],
        [0] * len(city_foss_participants),  # Z-axis base set to 0
        0.4, 0.4,  # Bar width and depth
        city_foss_participants['Participants'],  # Bar height based on participants
        shade=True
    )

    ax.set_xlabel('City')
    ax.set_ylabel('FOSS Type')
    ax.set_zlabel('Number of Participants')

    ax.set_xticks(list(city_map.values()))  # Set x-ticks to match city mapping
    ax.set_xticklabels(list(city_map.keys()), rotation=90)  # Rotate city labels for readability

    ax.set_yticks(list(foss_map.values()))  # Set y-ticks to match FOSS mapping
    ax.set_yticklabels(list(foss_map.keys()), rotation=90)  # Rotate FOSS labels for readability

    plt.title('3D Visualization of Workshops by City, FOSS, and Participants')
    plt.tight_layout()
    plt.show()


# Function to truncate sheet names to fit Excel's 31-character limit
def truncate_sheet_name(name):
    return name if len(name) <= 31 else name[:28] + '...'  # Truncate the sheet name if it exceeds 31 characters


# Base URL for scraping the workshop data
base_url = "https://spoken-tutorial.org/statistics/training/?training_planner__academic__state=&training_planner__academic__city=&training_planner__academic__institution_type=&training_planner__academic__institution_name=&department=&course_type=&course__foss=&sem_start_date_after=2022-01-01&sem_start_date_before=2023-01-01&lang=---------&status=1"

# FOSS types to filter by
foss_filter = ['Scilab', 'LaTeX', 'Linux', 'Python 3.4.3']
start_date = datetime.strptime('2022-01-01', '%Y-%m-%d')  # Start date for filtering workshops
end_date = datetime.strptime('2023-01-01', '%Y-%m-%d')  # End date for filtering workshops

# Scrape all data based on FOSS filters and date range
all_data = scrape_all_pages(base_url, foss_filter, start_date, end_date)

# Define columns for the DataFrame
columns = ['#', 'State', 'City', 'Institution', 'FOSS', 'Department', 'Type', 'Organiser', 'Date', 'Participants']

# Create a DataFrame and remove duplicates
df = pd.DataFrame(all_data, columns=columns).drop_duplicates()

# Write data to Excel and create visualizations
with pd.ExcelWriter('st_data.xlsx', mode='w') as writer:
    df.to_excel(writer, sheet_name='Workshops Data', index=False)  # Write workshop data to Excel

    foss_counts = piechart_visualization(df, writer)  # Create and save pie chart visualization
    city_counts = barchart_visualization(df, writer)  # Create and save bar chart visualization

    # Analyze data for a specific college
    college_name = "Shri Phanishwar Nath Renu Engineering College, Araria"
    foss_types, departments, organisers = analyze_college_data(df, college_name)

    if foss_types is not None:
        college_data = df[df['Institution'].str.contains(college_name, case=False, na=False)]  # Filter data for the college
        college_data_unique = college_data[['FOSS', 'Organiser', 'Date', 'Participants']].drop_duplicates()  # Get unique data
        college_data_unique.to_excel(writer, sheet_name=truncate_sheet_name(f'{college_name} Workshops'), index=False)  # Save to Excel

# Generate a 3D visualization of workshops
three_d_visualization(df)

print("Data has been saved to 'st_data.xlsx' and the visualizations have been displayed")
